package org.apache.pdfbox.multipdf;

import java.awt.geom.AffineTransform;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;

import org.apache.pdfbox.cos.COSArray;
import org.apache.pdfbox.cos.COSBase;
import org.apache.pdfbox.cos.COSDictionary;
import org.apache.pdfbox.cos.COSName;
import org.apache.pdfbox.cos.COSObject;
import org.apache.pdfbox.cos.COSStream;
import org.apache.pdfbox.io.IOUtils;
import org.apache.pdfbox.multipdf.Overlay.Position;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.PDResources;
import org.apache.pdfbox.pdmodel.common.PDRectangle;
import org.apache.pdfbox.pdmodel.graphics.form.PDFormXObject;

/**
 * Low-level engine for adding overlays to a PDF document
 * 
 * @author aditsu
 */
public class OverlayEngine
{
    private final PDDocument doc;
    private List<COSBase> origContents;
    private List<PDResources> origResources;

    /**
     * Creates an overlay engine
     * 
     * @param doc the input PDF document to be overlayed; it will be modified in place
     */
    public OverlayEngine(final PDDocument doc)
    {
        this.doc = doc;
    }

    /**
     * Creates an {@link OverlayPage} for overlaying a given page onto the input document
     * 
     * @param page the page to overlay
     * @return the generated {@link OverlayPage}
     * @throws IOException
     */
    public OverlayPage getOverlayPage(final PDPage page) throws IOException
    {
        COSBase contents = page.getCOSObject().getDictionaryObject(COSName.CONTENTS);
        PDResources resources = page.getResources();
        if (resources == null)
        {
            resources = new PDResources();
        }
        return new OverlayPage(page.getMediaBox(), createContentStream(contents),
                resources.getCOSObject());
    }

    private COSStream createContentStream(COSBase contents) throws IOException
    {
        List<COSStream> contentStreams = createContentStreamList(contents);
        // concatenate streams
        COSStream concatStream = doc.getDocument().createCOSStream();
        OutputStream out = concatStream.createOutputStream(COSName.FLATE_DECODE);
        for (COSStream contentStream : contentStreams)
        {
            InputStream in = contentStream.createInputStream();
            IOUtils.copy(in, out);
            out.flush();
            in.close();
        }
        out.close();
        return concatStream;
    }

    private List<COSStream> createContentStreamList(COSBase contents) throws IOException
    {
        List<COSStream> contentStreams = new ArrayList<COSStream>();
        if (contents instanceof COSStream)
        {
            contentStreams.add((COSStream) contents);
        }
        else if (contents instanceof COSArray)
        {
            for (COSBase item : (COSArray) contents)
            {
                contentStreams.addAll(createContentStreamList(item));
            }
        }
        else if (contents instanceof COSObject)
        {
            contentStreams.addAll(createContentStreamList(((COSObject) contents).getObject()));
        }
        else
        {
            throw new IOException("Contents are unknown type:" + contents.getClass().getName());
        }
        return contentStreams;
    }

    /**
     * Saves the current version of the document in memory, allowing it to be restored at a later
     * point
     */
    public void save()
    {
        final int n = doc.getNumberOfPages();
        origContents = new ArrayList<COSBase>(n);
        origResources = new ArrayList<PDResources>(n);
        for (int i = 0; i < n; ++i)
        {
            final PDPage page = doc.getPage(i);
            origContents.add(page.getCOSObject().getDictionaryObject(COSName.CONTENTS));
            origResources.add(page.getResources());
        }
    }

    /**
     * Restores a page in the input document to its last saved version
     * 
     * @param no the page number (0-based)
     */
    public void resetPage(final int no)
    {
        if (origContents == null) {
            throw new IllegalStateException("The document has not been saved");
        }
        final PDPage page = doc.getPage(no);
        page.getCOSObject().setItem(COSName.CONTENTS, origContents.get(no));
        page.setResources(origResources.get(no));
    }

    /**
     * Restores the input document to its last saved version
     */
    public void resetDocument()
    {
        for (int i = 0; i < doc.getNumberOfPages(); ++i)
        {
            resetPage(i);
        }
    }

    /**
     * Adds an overlay to a page in the input document
     * 
     * @param no the page number in the input document (0-based)
     * @param position the overlay position (foreground/background)
     * @param overlayPage the overlay to add, see {@link #getOverlayPage(PDPage)}
     * @param at the affine transform to use for the overlay
     * @throws IOException
     */
    public void processPage(final int no, final Position position, final OverlayPage overlayPage,
            final AffineTransform at) throws IOException
    {
        processPage(doc.getPage(no), position, overlayPage, at);
    }

    /**
     * Adds an overlay to a page in the input document
     * 
     * @param page the page to be overlayed (should be from the input document)
     * @param position the overlay position (foreground/background)
     * @param overlayPage the overlay to add, see {@link #getOverlayPage(PDPage)}
     * @param at the affine transform to use for the overlay
     * @throws IOException
     */
    public void processPage(final PDPage page, final Position position, final OverlayPage overlayPage,
            final AffineTransform at) throws IOException
    {
        COSDictionary pageDictionary = page.getCOSObject();
        COSBase originalContent = pageDictionary.getDictionaryObject(COSName.CONTENTS);
        COSArray newContentArray = new COSArray();
        switch (position)
        {
            case FOREGROUND:
                // save state
                newContentArray.add(createStream("q\n"));
                addOriginalContent(originalContent, newContentArray);
                // restore state
                newContentArray.add(createStream("Q\n"));
                // overlay content last
                addOverlay(page, overlayPage, at, newContentArray);
                break;
            case BACKGROUND:
                // overlay content first
                addOverlay(page, overlayPage, at, newContentArray);

                addOriginalContent(originalContent, newContentArray);
                break;
            default:
                throw new IOException("Unknown type of position:" + position);
        }
        pageDictionary.setItem(COSName.CONTENTS, newContentArray);
    }

    private static void addOriginalContent(COSBase contents, COSArray contentArray)
            throws IOException
    {
        if (contents instanceof COSStream)
        {
            contentArray.add(contents);
        }
        else if (contents instanceof COSArray)
        {
            contentArray.addAll((COSArray) contents);
        }
        else
        {
            throw new IOException("Unknown content type: " +
                    (contents == null ? "(null)" : contents.getClass().getName()));
        }
    }

    private void addOverlay(final PDPage page, final OverlayPage overlayPage,
            final AffineTransform at, final COSArray array) throws IOException
    {
        PDResources resources = page.getResources();
        if (resources == null)
        {
            resources = new PDResources();
            page.setResources(resources);
        }
        COSName xObjectId = createOverlayXObject(page, overlayPage,
                overlayPage.overlayContentStream);
        array.add(createOverlayStream(xObjectId, at));
    }

    private static COSName createOverlayXObject(PDPage page, OverlayPage overlayPage,
            COSStream contentStream)
    {
        PDFormXObject xobjForm = new PDFormXObject(contentStream);
        xobjForm.setResources(new PDResources(overlayPage.overlayResources));
        xobjForm.setFormType(1);
        xobjForm.setBBox(overlayPage.overlayMediaBox.createRetranslatedRectangle());
        xobjForm.setMatrix(new AffineTransform());
        PDResources resources = page.getResources();
        return resources.add(xobjForm, "OL");
    }

    private COSStream createOverlayStream(final COSName xObjectId, final AffineTransform at)
            throws IOException
    {
        // create a new content stream that executes the XObject content
        StringBuilder overlayStream = new StringBuilder();
        overlayStream.append("q\nq\n");
        double[] flatmatrix = new double[6];
        at.getMatrix(flatmatrix);
        for (double v : flatmatrix)
        {
            overlayStream.append(float2String((float) v));
            overlayStream.append(" ");
        }
        overlayStream.append(" cm\n/");
        overlayStream.append(xObjectId.getName());
        overlayStream.append(" Do Q\nQ\n");
        return createStream(overlayStream.toString());
    }

    /**
     * Calculates an affine transform that centers an overlay on a page
     * 
     * @param pageMediaBox the media box of the page that will get the overlay
     * @param overlayMediaBox the media box of the overlay
     * @return the affine transform to be used
     */
    public static AffineTransform getCenterTransform(final PDRectangle pageMediaBox,
            final PDRectangle overlayMediaBox)
    {
        AffineTransform at = new AffineTransform();
        float hShift = (pageMediaBox.getWidth() - overlayMediaBox.getWidth()) / 2.0f;
        float vShift = (pageMediaBox.getHeight() - overlayMediaBox.getHeight()) / 2.0f;
        at.translate(hShift, vShift);
        return at;
    }

    private static String float2String(float floatValue)
    {
        // use a BigDecimal as intermediate state to avoid
        // a floating point string representation of the float value
        BigDecimal value = new BigDecimal(String.valueOf(floatValue));
        String stringValue = value.toPlainString();
        // remove fraction digit "0" only
        if (stringValue.indexOf('.') > -1 && !stringValue.endsWith(".0"))
        {
            while (stringValue.endsWith("0") && !stringValue.endsWith(".0"))
            {
                stringValue = stringValue.substring(0, stringValue.length() - 1);
            }
        }
        return stringValue;
    }

    private COSStream createStream(String content) throws IOException
    {
        COSStream stream = doc.getDocument().createCOSStream();
        OutputStream out = stream.createOutputStream(
        		content.length() > 20 ? COSName.FLATE_DECODE : null);
        out.write(content.getBytes("ISO-8859-1"));
        out.close();
        return stream;
    }
}
